from tfgraphutils.util import *
from tensorflow import GraphDef

""" Internal Functions """
def _copy_node_to_graph_remove_inputs(node, new_graph):
    # Copy a node to a new graph, but remove inputs.
    new_node = new_graph.node.add()
    new_node.CopyFrom(node)
    new_node.ClearField("input")
    return new_node

""" Graph Manipulation Functions """
# TODO: Maybe let this function arbitrarily collapse any group of nodes.
# Will require more work on user end to collapse multiple namespaces if
# implemented this way.
def collapse_tf_graphdef_namespaces(tf_graphdef, namespace_plugin_map, nodes_to_exclude=[]):
    # Given tensorflow GraphDef and a dict of namespace names -> plugin names,
    # collapses those namespaces into single nodes representing plugins, excluding
    # those nodes specified in nodes_to_exclude.

    # Python 2.7 work-around for nonlocal variables
    class context(object):
        plugin_dummy_nodes = {}
        exclude_node_names = set(get_names_from_nodes(nodes_to_exclude))

    def get_reference_plugin_node(node_name):
        # Get the default plugin node provided by the user, or return None if this
        # does not belong in a plugin.

        def should_exclude_node(node_name):
            return node_name in context.exclude_node_names

        if should_exclude_node(node_name):
            # Don't put this node into a plugin, treat as normal node instead.
            return None
        # Check if this node should be omitted from the main graph and return the
        # name of a plugin if so. Try to match to the longest possible namespace path.
        best_match_depth = -1
        best_match = None
        for namespace in namespace_plugin_map:
            # Find the end point of the namespace
            current_depth = len(namespace)
            if namespace == node_name[:current_depth] and current_depth > best_match_depth:
                best_match_depth = current_depth
                best_match = namespace_plugin_map[namespace]
        return best_match

    def copy_node_to_collapsed_graph(node, tf_graphdef_collapsed):
        # Add a node, in the form of either a normal node or a
        # plugin dummy node to the collapsed graph.

        def get_plugin_dummy_node(reference_plugin_node):
            # Return a dummy node, representing a plugin, from the collapsed graph.
            if reference_plugin_node.name not in context.plugin_dummy_nodes:
                # Generates a new dummy node based on the reference node.
                context.plugin_dummy_nodes[reference_plugin_node.name] = tf_graphdef_collapsed.node.add()
                context.plugin_dummy_nodes[reference_plugin_node.name].CopyFrom(reference_plugin_node)
                if reference_plugin_node.op is None or reference_plugin_node.op == "":
                    context.plugin_dummy_nodes[reference_plugin_node.name].op = reference_plugin_node
            return context.plugin_dummy_nodes[reference_plugin_node.name]

        reference_plugin_node = get_reference_plugin_node(node.name)
        if reference_plugin_node:
            # If node needs to be collapsed, get a plugin dummy node
            new_node = get_plugin_dummy_node(reference_plugin_node)
        else:
            # otherwise, copy the node over without its inputs
            new_node = _copy_node_to_graph_remove_inputs(node, tf_graphdef_collapsed)
        return new_node, reference_plugin_node

    def add_collapsed_inputs_to_node(node, collapsed_node, reference_plugin_node):
        # Add inputs from node to collapsed_node, collapsing any inputs that are
        # in a plugin along the way.
        num_inputs_added = 0
        for input_name in node.input:
            # We don't care if this is a control input
            input_name = input_name.replace('^', '')
            # Figure out which plugin this input belongs to.
            input_reference_plugin_node = get_reference_plugin_node(input_name)
            # If the input is in a plugin, use the plugin name rather than the input name.
            input_name = input_reference_plugin_node.name if input_reference_plugin_node else input_name
            # Add this input as long it is not in the same plugin.
            if not input_reference_plugin_node or input_reference_plugin_node != reference_plugin_node:
                collapsed_node.input.append(input_name)
                num_inputs_added += 1
        return num_inputs_added

    def generate_new_graph():
        tf_graphdef_collapsed = GraphDef()
        # Main Loop
        for node in tf_graphdef.node:
            new_node, reference_plugin_node = copy_node_to_collapsed_graph(node, tf_graphdef_collapsed)
            # Update inputs to match new, collapsed nodes
            add_collapsed_inputs_to_node(node, new_node, reference_plugin_node)
        # Done!
        return tf_graphdef_collapsed

    return generate_new_graph()

def remove_nodes_from_graph_by_name(tf_graphdef, removable_names=set([])):
    # Given a graph def, removes nodes corresponding to the names provided and
    # returns a new GraphDef. Does not forward inputs.

    def should_remove_node_name(node_name):
        # Determine whether this node_name should be removed from the graph
        node_name = node_name.replace('^', '')
        should_remove_node = node_name in removable_names
        # Check if this node shows up as a control dependency
        should_remove_control_dependency = '^' + node_name in removable_names
        return should_remove_node or should_remove_control_dependency

    def add_unremoved_inputs_to_node(node, new_node):
        # Copy inputs from node to new_node, excluding removed ones
        num_inputs_added = 0
        for input_name in node.input:
            if not should_remove_node_name(input_name):
                # Add inputs only when they are not supposed to be removed
                num_inputs_added += 1
                new_node.input.append(input_name)
        return num_inputs_added

    def generate_new_graph():
        pruned_tf_graphdef = GraphDef()
        for node in tf_graphdef.node:
            if not should_remove_node_name(node.name):
                # If this node should not be removed, add it to the graph.
                new_node = _copy_node_to_graph_remove_inputs(node, pruned_tf_graphdef)
                # Copy over those inputs which haven't been removed
                add_unremoved_inputs_to_node(node, new_node)
        # Done!
        return pruned_tf_graphdef

    # Generate the new graph.
    return generate_new_graph()

def remove_nodes_from_graph(tf_graphdef, removable_nodes):
    # Allows for removal of nodes based on node references directly.
    removable_names = get_names_from_nodes(removable_nodes)
    return remove_nodes_from_graph_by_name(tf_graphdef, removable_names)

# FIXME: Handle control inputs properly when bridging.
# Figure out the duplicate input situation.
def bridge_nodes_in_graph_by_name(tf_graphdef, bridge_names=set([])):
    """
    Given a graphdef and a container of node names, generates a new graph
        with all the inputs of the specified nodes recursively forwarded, and the nodes
        themselves removed.
    WARNING: Remove asserts before attempting to bridge.
    ASSERTS WILL NOT BE FORWARDED!
    """
    # Python 2.7 work-around for nonlocal variables
    class context(object):
        output_map = get_output_map(tf_graphdef)

    def should_bridge_node(node):
        # Bridge if the node is in the list...
        is_in_bridge_names = node.name in bridge_names
        # ...unless it has control edge inputs
        has_control_edge = False
        for input_name in node.input:
            if '^' in input_name:
                has_control_edge = True
        # Finally, in order to be forwarded it must have outputs
        has_outputs = node.name in context.output_map and len(context.output_map[node.name]) != 0
        return has_outputs and is_in_bridge_names and not has_control_edge

    def generate_shallow_input_replacements():
        shallow_input_replacements = defaultdict(list)
        # Traverse the graph once to get a shallow mapping of input -> replacements
        for node in tf_graphdef.node:
            if should_bridge_node(node):
                shallow_input_replacements[node.name].extend(node.input)
        return shallow_input_replacements

    def update_shallow_input_replacements(shallow_input_replacements):
        # Traverse the input replacement map and generate a map of true input replacements.
        for node_name in shallow_input_replacements:
            for input_name in shallow_input_replacements[node_name]:
                if input_name in shallow_input_replacements:
                    # Append replacements to the end of the input list
                    shallow_input_replacements[node_name].extend(shallow_input_replacements[input_name])
                    # Pop replaced inputs from the front.
                    shallow_input_replacements[node_name].remove(input_name)
        # Done!
        return shallow_input_replacements

    def add_bridged_inputs_to_node(node, new_node, true_input_replacements):
        # Copy inputs from node to new_node, bridging those which need to be.

        def should_replace_input(input_name):
            # REVIEW: Might need to do this a different way later.
            # Check the true input name, not just as a control input.
            # new_input_name = input_name.replace('^', '')
            # if new_input_name in true_input_replacements:
            #     return new_input_name
            # return None
            return input_name.replace('^', '') in true_input_replacements

        num_inputs_added = 0
        for input_name in node.input:
            # REVIEW: Might need to do this a different way later.
            # new_input_name = get_input_name_to_replace(input_name)
            # if new_input_name:
            #     if '^' in input_name:
            #             logger.debug("Replacing " + input_name + " with " + str(['^' + rep for rep in true_input_replacements[new_input_name]]))
            #         # If the original input was a control input, add replacements as control inputs.
            #         new_node.input.extend(['^' + rep for rep in true_input_replacements[new_input_name]])
            #     else:
            #             logger.debug("Replacing " + input_name + " with " + str(true_input_replacements[input_name]))
            #         # Otherwise add as normal.
            #         new_node.input.extend(true_input_replacements[new_input_name])
            if should_replace_input(input_name):
                # REVIEW: Do we need to check for unique inputs here?
                # unique_replacement_names = [replacement_name
                #     for replacement_name in true_input_replacements[input_name]
                #         if replacement_name not in new_node.input]
                unique_replacement_names = true_input_replacements[input_name]
                new_node.input.extend(unique_replacement_names)
                num_inputs_added += len(unique_replacement_names)
            else:
                new_node.input.append(input_name)
                num_inputs_added += 1
        return num_inputs_added

    def generate_new_graph(true_input_replacements):
        # Traverse the graph and replace nodes
        bridged_tf_graphdef = GraphDef()
        for node in tf_graphdef.node:
            if 'Preprocessor/map/while/Switch:1' in node.name:
                print(node)
            if not should_bridge_node(node):
                # If this node should not be bridged, add it to the graph
                new_node = _copy_node_to_graph_remove_inputs(node, bridged_tf_graphdef)
                # Add inputs which haven't been bridged, otherwise add input replacements
                add_bridged_inputs_to_node(node, new_node, true_input_replacements)
        # Done!
        return bridged_tf_graphdef

    # Initial pass to get 1-layer deep replacements.
    shallow_input_replacements = generate_shallow_input_replacements()
    # Update shallow_input_replacements to be true replacements.
    true_input_replacements = update_shallow_input_replacements(shallow_input_replacements)
    # Generate the new graph.
    return generate_new_graph(true_input_replacements)

def bridge_nodes_in_graph(tf_graphdef, bridge_nodes):
    # Allows for removal of nodes based on node references directly.
    bridge_names = get_names_from_nodes(bridge_nodes)
    return bridge_nodes_in_graph_by_name(tf_graphdef, bridge_names)
