from tfgraphutils.util import *
import re

""" Graph Search Functions """
def find_nodes_by_ops(tf_graphdef, ops=set([])):
    # Given tf_graphdef and a set of ops, generates a list of nodes.
    return [node for node in tf_graphdef.node if node.op in ops]

def find_nodes_by_names(tf_graphdef, names=set([])):
    # Given tf_graphdef and a set of names, generates a list of nodes.
    # Names can be RegExs.

    def has_name(node):
        for name in names:
            # If the name is present in the node name, this is a match!
            matches_name = name in node.name or name == node.name
            matches_regex = re.search(name, node.name)
            if matches_name or matches_regex:
                return True
        return False

    return [node for node in tf_graphdef.node if has_name(node)]

def find_nodes_by_paths(tf_graphdef, paths=set([])):
    # Matches exact paths. Paths can use RegExs.

    def has_path(node):
        for path in paths:
            # If the namespace path is present in the node path,
            # the node is in the namespace!
            matches_path = path == node.name
            matches_regex = re.match(path, node.name)
            if matches_path or matches_regex:
                return True
        return False

    return [node for node in tf_graphdef.node if has_path(node)]

def find_unused_nodes_on_removal(tf_graphdef, node_removal_list):
    # Given a list of nodes to remove, return a list of ALL
    # nodes that are now unnecessary in the graph.

    # Python 2.7 work-around for nonlocal variables
    class context(object):
        node_name_map = get_node_names_map(tf_graphdef)
        # Since output_map will be modified, need a local copy
        output_map = get_output_map(tf_graphdef)

    def recursively_remove_inputs(node):
        # Given one node, return a set containing it and all its hanging inputs
        removable_nodes_list = [node]
        for input_name in node.input:
            # Remove this node from the output of it's inputs
            if input_name in context.output_map and node.name in context.output_map[input_name]:
                context.output_map[input_name].remove(node.name)
            # Recursively remove any inputs which are left hanging
            if len(context.output_map[input_name]) == 0:
                input_node = context.node_name_map[input_name]
                removable_nodes_list.extend(recursively_remove_inputs(input_node))
        return removable_nodes_list

    # Nodes that can be removed based on nodes going to be removed.
    removable_nodes_list = []
    for node in node_removal_list:
        removable_nodes_list.extend(recursively_remove_inputs(node))
    return removable_nodes_list

def find_output_nodes_of_graph(tf_graphdef):
    # Find the outputs of the graph i.e. any nodes
    # which do not serve as inputs to other nodes.
    output_map = get_output_map(tf_graphdef)
    graph_outputs = []
    for node in tf_graphdef.node:
        if node.name not in output_map:
            graph_outputs.append(node)
    return graph_outputs
